import os


def detect_ec2(ctx):
    """
    detects whether the current convocation is running in ec2
    by checking the /sys/hypervisor/uuid file
    """
    if os.path.exists('/sys/hypervisor/uuid'):
        result = ctx.run('head -c 3 /sys/hypervisor/uuid')
        return result.stdout.startswith('ec2')
    return False



from raft import task
from raft.collection import Collection


def render(env, source, destination, context=None):
    """
    :param jinja2.Environment env:
    :param str source:
    :param str destination:
    :param dict[Any, Any] context:
    """
    t = env.get_template(source)
    context = context or {}
    body = t.render(**context)
    with open(destination, 'w') as f:
        f.write(body)


@task(help=dict(
    name='the name of the project',
    path='the parent directory in which {name} will be created',
    source_dir='the name of the source directory within our project, for '
               'most projects, leave this blank so that the value defaults '
               'to {name}',
    skip_setup='if this flag is specified, we will not create a virtual '
               'environment and install base requirements as part of '
               'kicking off the new directory',
))
def template(ctx, name, path='.', source_dir=None, skip_setup=False):
    """
    creates a template for a pypi project using our standard defaults
    """
    from ..base.projects import new_project
    boo = not skip_setup
    templates = [
        '.bumpversion.cfg',
        '.coveragerc',
        '.gitignore',
        'dev.txt',
        'Makefile',
        'MANIFEST.in',
        'pytest.ini',
        'README.md',
        'requirements.txt',
        'setup.py',
        'sesame_meow_cat/__init__.py',
        'sesame_meow_cat/version.txt',
        'test/__init__.py',
        'test/conftest.py',
    ]
    new_project(
        ctx, name, path, source_dir, boo,
        templates, package='convocations.pypi')


ns = Collection(template)



from raft import task
from ...base.utils import print_table
from ..base import AwsTask


@task(klass=AwsTask)
def rds_instances(ctx, session=None, profile=None, **kwargs):
    """
    shows all rds instances and their endpoints
    """
    rds = session.client('rds')
    instances = rds.describe_db_instances()
    instances = instances['DBInstances']
    header = [ 'name', 'engine', 'endpoint', 'status', ]
    rows = []
    for x in instances:
        rows.append([
            x['DBInstanceIdentifier'],
            f"{x['Engine']} {x['EngineVersion']}",
            f"{x['Endpoint']['Address']}:{x['Endpoint']['Port']}",
            x['DBInstanceStatus'],
        ])
    rows.sort(key=lambda lx: lx[0])
    print_table(header, rows)




from raft.tasks import task
from ...base.utils import print_table
from ..base import AwsTask, yielder

__all__ = [
    'search_s3',
    'add_name_tag_to_bucket',
]


@task(klass=AwsTask)
def search_s3(bucket, prefix, session=None, **kwargs):
    """
    searches the bucket at the given prefix, recursively, and prints
    matching content
    """
    s3 = session.client('s3')
    method = 'list_objects_v2'
    suffixes = [ 'b', 'kb', 'mb', 'gb', 'tb' ]
    config = dict(
        Bucket=bucket,
        Prefix=prefix,
        PaginationConfig={
            'PageSize': 1000,
        })
    header = [ 'key', 'last_modified', 'file_size' ]
    rows = []
    for x in yielder(s3, method, session, **config):
        file_size = x['Size']
        n_suffix = 0
        while file_size > 1024 and n_suffix < len(suffixes):
            n_suffix += 1
            file_size /= 1024
        rows.append([
            x['Key'],
            x['LastModified'].isoformat(),
            f"{file_size:.2f}{suffixes[n_suffix]}"
        ])
    print_table(header, rows)


@task(klass=AwsTask)
def add_name_tag_to_bucket(ctx, for_reals=False, session=None, profile=None, **kwargs):
    """
    adds a name tag to all s3 buckets with the same name as the bucket name
    """
    from botocore.exceptions import ClientError
    s3 = session.client('s3')
    response = s3.list_buckets()
    for x in response['Buckets']:
        bucket = x['Name']
        try:
            tagging = s3.get_bucket_tagging(Bucket=bucket)
            tags = { tag['Key']: tag['Value'] for tag in tagging['TagSet'] }
        except ClientError:
            tags = {}
        if 'Name' not in tags:
            print(f'[{bucket}] adding Name tag')
            tags['Name'] = bucket
            tag_set = [ dict(Key=key, Value=value) for key, value in tags.items() ]
            if for_reals:
                s3.put_bucket_tagging(
                    Bucket=bucket,
                    Tagging=dict(TagSet=tag_set))



from raft.tasks import task
from convocations.base.utils import print_table, notice, notice_end
from .base import CheckpointTask


@task(klass=CheckpointTask)
def group(ctx, name, client=None, host=None, **kwargs):
    """
    shows information about the network group `named`
    """
    from cpapi.api_response import APIResponse
    notice('getting group')
    response: APIResponse = client.api_call('show-group', payload={
        'name': name,
    })
    data = response.data
    notice_end(data.get('name') or data.get('message'))
    members = [ [ x['name'] ] for x in data['members'] ]
    print_table([ 'name' ], members)
    return data



from datetime import datetime
import os
from tempfile import NamedTemporaryFile

from raft import task
from convocations.aws.base import AwsTask


__all__ = [
    'aws_apt_sources',
    'migrate',
    'setup',
    'setup_galaxy',
    'bundle',
    'deploy',
    'Codebuild',
]


@task
def setup(ctx):
    """
    installs dependencies in the codebuild environment
    this installation is dead simple because we don't have
    to worry about python virtual environments
    """
    pip = 'pip install -q'
    ctx.run(f'{pip} -U wheel')
    ctx.run(f'{pip} -U pip setuptools nodeenv')
    ctx.run(f'{pip} -r requirements.txt')


@task
def setup_galaxy(ctx):
    """
    installs ansible-galaxy dependencies
    """
    ctx.run('ansible-galaxy install -r deploy/requirements.yml')


@task
def client(ctx):
    """
    builds the vue.js client
    """
    ctx.run('npm i -g parcel-bundler')
    with ctx.cd('client'):
        ctx.run('npm i')
        ctx.run('npm run build')


@task
def aws_apt_sources(ctx):
    """Fixes /etc/apt/sources.list by pointing to aws ubuntu mirrors"""
    with open('/etc/apt/sources.list', 'r') as f:
        data = f.read()
    region = os.environ.get('AWS_DEFAULT_REGION', 'us-east-2')
    data = data.replace(
        'http://archive.ubuntu.com', f'http://{region}.ec2.archive.ubuntu.com'
    )
    with open('/etc/apt/sources.list', 'w') as f:
        f.write(data)


@task
def migrate(ctx):
    """
    runs django migrations from codebuild
    -- not for use on local
    ```yaml
    codebuild:
      envs:
        master: prod
        staging: staging
    ```
    """
    from .utils import which_python
    envs = ctx.codebuild
    migrating_branches = list(envs.keys())
    x = Codebuild()
    if x.is_branch:
        branch = x.git_branch
    else:
        print('No migrations, we are not in a branch codebuild')
        return

    if branch not in migrating_branches:
        print(f'only migrating for branches {migrating_branches}')
        return

    python = which_python(ctx)
    tundra_env = envs[branch]
    os.environ['TUNDRA_ENV'] = tundra_env
    ctx.run(f'{python} manage.py migrate')


@task
def bundle(ctx):
    """
    creates a bundle and deploys to s3 as an artifact for use with deploys
    requires the following configuration:

    ```yaml
    codebuild:
      bundles:
        versioned:
          name: path/to/{version}
          bucket: s3.bucket.name
        staging:
          name: path/to/staging
          bucket: s3.bucket.name
        master:
          name: path/to/prod
          bucket: s3.bucket.name

    the bundles that we create will be stored at, for example, path/to/prod/latest.zip
      and archive/path/to/prod/<dated name>.zip
    ```
    """
    import pytz
    from ..base.utils import get_context_value
    from ..base.utils import notice, notice_end
    x = Codebuild()
    conf = get_context_value(ctx, 'codebuild.bundles') or {}
    if x.is_branch:
        conf = conf[x.git_branch]
        version = x.git_branch
    else:
        conf = conf.versioned
        version = x.git_version
    name = conf.name.format(version=version)
    bucket = conf.bucket
    t = NamedTemporaryFile(delete=False)  # pylint: disable=consider-using-with
    notice(f'creating {t.name}.zip')
    ctx.run(f"zip -q -r {t.name}.zip . -x '/data/*'")
    notice_end()
    latest = 'latest.zip'
    latest = f's3://{bucket}/{name}/{latest}'
    dated = datetime.now(pytz.timezone('CST6CDT'))
    dated = dated.strftime('%Y%m%d.%H%M.zip')
    dated = f's3://{bucket}/archive/{name}/{dated}'
    notice(f'uploading {t.name}.zip => {latest}')
    ctx.run(f'aws s3 cp {t.name}.zip {latest}', hide='both')
    notice_end()
    notice(f'uploading {dated}')
    ctx.run(f'aws s3 cp {latest} {dated}', hide='both')
    notice_end()
    os.remove(f'{t.name}.zip')


@task(klass=AwsTask)
def deploy(ctx, profile=None, session=None, **kwargs):
    """
    deploys from codebuild to the specified tundra environment

    aws:
      profile: profile_name
    tundra:
      deploy:
        deploy_keys:
          limit1: namespace/key1
          limit2: namespace/key2
        staging:
          limit: staging_limit
          branch: staging
        prod:
          limit: production_limit
          branch: master
    """
    from .deploy import _deploy
    x = Codebuild()
    branch = x.git_branch
    conf = ctx.tundra.deploy
    found = False
    which = None
    for x, value in conf.items():
        if value['branch'] == branch:
            found = True
            which = x
    if not branch or not found:
        print(f'Nothing to be done, we do not deploy {branch}')
        return
    conf = conf.get(which)
    limit = conf['limit']
    _deploy(ctx, conf, limit, branch, session=session)


class Codebuild:
    ref = None
    git_branch = None
    git_version = None

    @property
    def is_tag(self):
        return self.ref.startswith('tag/')

    @property
    def is_branch(self):
        return bool(self.git_branch)

    def __init__(self):
        self.ref = os.environ.get('CODEBUILD_WEBHOOK_TRIGGER', '')
        self.git_branch = self.ref.rsplit('/', 1)[-1].strip()
        if self.is_tag and self.git_branch.startswith('v'):
            self.git_version = self.git_branch
            self.git_branch = None
        if self.git_branch == '':
            self.git_branch = os.environ.get('CODEBUILD_SOURCE_VERSION', '')



from tempfile import NamedTemporaryFile
from uuid import uuid4

from raft import task
from .base import XmlApiTask


@task(klass=XmlApiTask)
def install_cert(ctx, name, cert, key, **kwargs):
    """
    installs a cert to the firewall
    """
    from lxml import etree
    from OpenSSL.crypto import dump_privatekey
    from OpenSSL.crypto import load_privatekey
    from OpenSSL.crypto import FILETYPE_PEM
    api = kwargs['api']
    passphrase = uuid4().hex
    with open(key, 'rb') as f:
        x509_key = load_privatekey(FILETYPE_PEM, f.read())
    with NamedTemporaryFile('wb', buffering=1) as f:
        st = dump_privatekey(
            FILETYPE_PEM,
            x509_key,
            passphrase=passphrase.encode())
        f.write(st)
        f.flush()

        response = api.import_cert(name, 'certificate', cert)
        print(etree.tostring(response, pretty_print=True).decode())

        response = api.import_cert(
            name, 'private-key', f.name, passphrase=passphrase)
        print(etree.tostring(response, pretty_print=True).decode())



from typing import Iterable
from typing import List

from raft import task


@task(help=dict(perms='read, write, or manage'), iterable=['perms'])
def grant_access(ctx, site_name, app_name, perms=None):
    """
    grants the app with name app_name read and write permissions to the site_name sharepoint site

    this convocation is useful when you have created an azure application
    registration with api permission Sites.Selected and need to authorize
    the application to connect to a particular sharepoint site.
    """
    from msal import ConfidentialClientApplication
    from office365.graph_client import GraphClient
    from office365.directory.applications.application import Application
    from office365.sharepoint.sites.site import Site
    perms = perms or [ 'read', 'write' ]
    app = ConfidentialClientApplication(
        authority=f'https://login.microsoftonline.com/{ctx.tenant_id}',
        client_id=ctx.client_id,
        client_credential=ctx.client_secret,
    )
    sharepoint_base_url = f'https://{ctx.tenant_name}.sharepoint.com'
    token_json = app.acquire_token_for_client(scopes=[
        # f'{sharepoint_base_url}/.default',
        'https://graph.microsoft.com/.default',
    ])
    site_url = f'{sharepoint_base_url}/sites/{site_name}'
    session = GraphClient(lambda: token_json)
    result: Iterable[Application] = session.applications.get().execute_query()
    app_id = None
    for x in result:
        if x.display_name.lower() == app_name.lower():
            app_id = x.id
            break
    if not app_id:
        print(f'could not find {app_name} in azure application directory')
    print(f'{app_name} => {app_id}')
    site: Site = session.sites.get_by_url(site_url).execute_query()
    print(f'{site.name} => {site.id}')
    result = site.permissions.add(
        roles=perms,
        grantedToIdentities=[{
            'application': {
                'id': app_id,
                'displayName': app_name,
            }
        }],
    ).execute_query()
    print(result.roles)


@task
def check_access(ctx, site_name, app_name):
    """
    checks the permissions for the app with name app_name
    on the site_name sharepoint site

    this convocation is useful when you have created an azure application
    registration with api permission Sites.Selected and need to authorize
    the application to connect to a particular sharepoint site.
    """
    from msal import ConfidentialClientApplication
    from office365.graph_client import GraphClient
    from office365.directory.applications.application import Application
    from office365.sharepoint.sites.site import Site
    from office365.onedrive.permissions.permission import Permission
    from office365.directory.identities.identity_set import IdentitySet
    app = ConfidentialClientApplication(
        authority=f'https://login.microsoftonline.com/{ctx.tenant_id}',
        client_id=ctx.client_id,
        client_credential=ctx.client_secret,
    )
    sharepoint_base_url = f'https://{ctx.tenant_name}.sharepoint.com'
    token_json = app.acquire_token_for_client(scopes=[
        # f'{sharepoint_base_url}/.default',
        'https://graph.microsoft.com/.default',
    ])
    site_url = f'{sharepoint_base_url}/sites/{site_name}'
    session = GraphClient(lambda: token_json)
    result: Iterable[Application] = session.applications.get().execute_query()
    app_id = None
    for x in result:
        if x.display_name.lower() == app_name.lower():
            app_id = x.id
            break
    if not app_id:
        print(f'could not find {app_name} in azure application directory')
    print(f'{app_name} => {app_id}')
    site: Site = session.sites.get_by_url(site_url).execute_query()
    print(f'{site.name} => {site.id}')
    result: List[Permission] = site.permissions.get().execute_query()
    for x in result:
        p: Permission = x.get().execute_query()
        sets: List[IdentitySet] = p.granted_to_identities
        for identity_set in sets:
            print(
                p.roles,
                identity_set.application.id,
                identity_set.application.displayName)



import os.path
from raft import task
from ..aws.base import AwsTask
from ..base.utils import get_context_value, notice, notice_end


@task
def test(ctx, f=None, quiet=False, **kwargs):
    """
    runs pytest against the specified directory with a default `-vvv` option

    ```yaml
    py:
      test_dir: tests
    ```
    :param f: the path to the test directory or to a pytest file
    :param quiet: if specified, changes pytest to regular output mode
    :param ctx:
    :param kwargs:
    """
    f = f or get_context_value(ctx, 'py.test_dir')
    if not f:
        for x in 'test', 'tests':
            if os.path.exists(x):
                f = x
    notice(f'looking for {f}')
    if not os.path.exists(f):
        notice_end(False)
        return
    notice_end()

    verbose = '' if quiet else '-vvv'
    ctx.run(
        'source bin/activate'
        f' && coverage run -m pytest {f} {verbose}'
        ' && coverage report --show-missing',
        pty=True
    )


@task(klass=AwsTask)
def test_with_aws(ctx, f=None, quiet=False, **kwargs):
    """
    runs pytest against the specified directory with a default `-vvv` option
    runs sso_login first in case you are running tests that require
    a waddle config

    ```yaml
    aws:
      profile: default
    py:
      test_dir: tests
    ```
    :param f: the path to the test directory or to a pytest file
    :param quiet: if specified, changes pytest to regular output mode
    :param ctx:
    :param kwargs:
    """
    return test(ctx, f, quiet, **kwargs)



import time

from raft import task


@task
def upgrade(ctx, version='10.1.5'):
    """
    Updates the ha pair
    """
    from panos.firewall import Firewall
    from .base import XmlApi
    from ..base.utils import notice, notice_end
    from lxml import etree
    username, password = ctx.palo_alto.username, ctx.palo_alto.password
    primary = Firewall(ctx.palo_alto.host, username, password)
    secondary = Firewall(ctx.palo_alto.peer, username, password)
    api = XmlApi(ctx.palo_alto.host, username, password)
    notice('setting preemptive to no')
    params = dict(
        xpath=(
            '/config/devices/entry[@name="localhost.localdomain"]'
            '/deviceconfig/high-availability/group/election-option'
        ),
        element='<preemptive>no</preemptive>',
    )
    r = api.post('config', 'set', **params)
    print(etree.tostring(r, pretty_print=True).decode('utf-8'))
    primary.commit(sync=True)
    notice_end()
    notice(f'upgrading secondary to version {version}')
    secondary.software.upgrade_to_version(version)
    notice_end()
    notice('waiting 120 seconds for secondary to come back up')
    time.sleep(120)
    notice_end()
    notice('checking version after upgrade')
    secondary.refresh_system_info()
    notice_end(secondary.version)
    input('please press <enter> to continue')
    notice('suspending ha on active peer')
    primary.op('request high-availability state suspend')
    notice_end()
    notice(f'upgrading primary to version {version}')
    primary.software.upgrade_to_version(version)
    notice_end()
    notice('waiting 120 seconds for primary to come back up')
    time.sleep(120)
    notice_end()
    notice('checking version after upgrade')
    primary.refresh_system_info()
    notice_end(primary.version)

    notice('setting preemptive to yes')
    params = dict(
        xpath=(
            '/config/devices/entry[@name="localhost.localdomain"]'
            '/deviceconfig/high-availability/group/election-option'
        ),
        element='<preemptive>yes</preemptive>',
    )
    r = api.post('config', 'set', **params)
    notice_end()
    notice('committing')
    primary.commit(True)
    notice_end()
    print(etree.tostring(r, pretty_print=True).decode('utf-8'))
    notice('turning on ha on primary')
    primary.op('request high-availability state functional')
    notice_end()
    notice('turning off ha on secondary to force failover')
    secondary.op('request high-availability state suspend')
    notice_end()
    time.sleep(10)
    notice('turning ha back on secondary')
    secondary.op('request high-availability state functional')
    notice_end()



import os
import shutil
import sys
from importlib.resources import files


def new_project(ctx, name, dir_name='.', source_dir=None,
                setup=False, jinja_templates=None, package=None):
    """
    all templates and filenames will have the text `sesame_meow_cat` replaced
    with the user-provided source_dir
    """
    from jinja2.environment import Environment
    from jinja2.loaders import PackageLoader
    current = os.getcwd()
    os.chdir(dir_name)
    os.makedirs(name, exist_ok=True)
    os.chdir(name)
    source_dir = source_dir or name
    meow_cat = 'sesame_meow_cat'
    env = Environment(loader=PackageLoader(package))
    templates = files(package).joinpath('templates')
    templates_dir = templates.as_posix()
    jinja_templates = [ x.replace(meow_cat, source_dir) for x in jinja_templates ]
    for x in templates.glob('**/*'):
        filename = x.relative_to(templates_dir).as_posix()
        print(filename)
        if '__pycache__' in filename or filename.endswith('.pyc'):
            continue
        template_name = filename
        filename = filename.replace(meow_cat, source_dir)
        if x.is_dir():
            print(f'creating dir {filename}')
            os.makedirs(filename, exist_ok=True)
            continue

        if template_name in jinja_templates:
            print(f'templating {filename}')
            t = env.get_template(template_name)
            contents = t.render(name=name)
            with open(filename, 'w') as f:
                f.write(contents)
        else:
            print(f'copying {x.as_posix()} => {filename}')
            shutil.copy2(x.as_posix(), filename)
    if setup and sys.platform == 'linux':
        ctx.run('make setup')
    os.chdir(current)



import time

from raft import task

from convocations.aws.base import AwsTask
from convocations.base.utils import notice, notice_end


@task(klass=AwsTask)
def task_logs(ctx, session, cluster_name, task_name, task_id):
    from ..cloudwatch import yield_logs
    logs = session.client('logs')
    ecs = session.client('ecs')
    start_time = None
    group_name = cluster_name
    stream_name = f'{cluster_name}/{task_name}/{task_id}'
    try:
        for n in range(0, 600):
            try:
                lines = list(
                    yield_logs(logs, group_name, stream_name, start_time)
                )
                lines.sort(key=lambda lx: lx['timestamp'])
                for x in lines:
                    print(x['message'])
                if lines:
                    start_time = lines[-1]['timestamp']
                if n % 5 == 4:
                    response = ecs.describe_tasks(
                        cluster=cluster_name, tasks=[task_id]
                    )
                    response = response['tasks'][0]
                    status = response['lastStatus']
                    print(f'task status: {status}')
                    if status == 'STOPPED':
                        break
                time.sleep(1)
            except logs.exceptions.ResourceNotFoundException:
                time.sleep(1)
    except KeyboardInterrupt:
        pass


@task(klass=AwsTask)
def get_latest_revision(ctx, task_name, session=None, quiet=False, **kwargs):
    """
    prints out the latest task revision for a task
    :param ctx:
    :param task_name: the name of the task
    :param session: the session passed in from `AwsTask`
    :param quiet: if this flag is set, output is repressed
    :param kwargs:
    :return:
    """
    talkative = not quiet
    if talkative:
        notice('getting task definition')
    ecs = session.client('ecs')
    result = ecs.describe_task_definition(taskDefinition=task_name)
    x = result['taskDefinition']
    result = f"{task_name}:{x['revision']}"
    if talkative:
        notice_end(result)
    return result




from raft import task
from ..base import AwsTask


@task(klass=AwsTask, help=dict(
    role='the name of the role to login to',
    force_sso='specify to bypass cache usage',
    profile='the already-configured aws profile to use for sso login',
))
def sso_login(ctx, role=None, force_sso=False, session=None, profile=None, **kwargs):
    """
    logs in via sso to the role name specified.  If no role name is specified,
    login will log into the first role alphabetically.  follows the boto3
    convention and caches short-term credentials in the ~/.aws/sso/cache
    directory.  Use the force_sso flag to bypass the use of cached credentials.
    """



import os
from raft import task
from convocations.aws.open_id import sso_login
from convocations.base.utils import notice, notice_end


@task
def download_drivers(ctx, profile=None, **kwargs):
    """
    downloads database drivers from s3 and install them

    ```yaml
    drivers:
      profiles:
        - profile1
        - profile2
      bucket: isos.example.com
      keys:
        - path/to/driver1.zip
        - path/to/driver2.tgz
    ```
    """
    from boto3 import Session
    session = Session()
    profiles = ctx.drivers.profiles
    bucket = ctx.drivers.bucket
    keys = ctx.drivers.keys
    if profile:
        profiles.insert(profile, 0)
    logged_in = set()
    d = './drivers'
    filenames = [ os.path.join(d, os.path.basename(key)) for key in keys ]
    os.makedirs(d, exist_ok=True)
    for x in profiles:
        if x not in session.available_profiles:
            continue
        try:
            if profile not in logged_in:
                sso_login(ctx, profile=profile)
                logged_in.add(profile)
            s3 = session.client('s3')
            for key, filename in zip(keys, filenames):
                notice(f'downloading {filename}')
                if not os.path.exists(filename):
                    s3.download_file(bucket, x, filename)
                    notice_end()
                else:
                    notice_end('already exists')
            break
        except:
            pass


@task
def build_base_image(ctx, profile=None):
    """
    builds the base docker image that we use with airflow.
    """
    download_drivers(ctx, profile)
    image_name = ctx.docker.image
    version = ctx.docker.version
    try:
        docker_file = ctx.docker.dockerfile
    except AttributeError:
        docker_file = 'Dockerfile'
    ctx.run(
        'docker build '
        f'  -t {image_name}:{version} '
        f'  -t {image_name}:latest '
        f'  -f {docker_file}'
        '  .'
    )



import time

from raft.tasks import task


@task
def wait_for_mysql(ctx, service_name='db'):
    from mysql.connector import connect
    from mysql.connector.errors import OperationalError
    import yaml
    with open('docker-compose.yml', 'r') as f:
        compose = yaml.load(f, Loader=yaml.SafeLoader)
        db_port = compose['services'][service_name]['ports'][0].split(':')[1]
    n_max = 40
    for x in range(1, n_max + 1):
        try:
            print(f'checking for mysql on port {db_port} attempt #{x}')
            conn = connect(host='127.0.0.1', user='root', password='', port=db_port)
            cursor = conn.cursor()
            cursor.execute('select now()')
            data = cursor.fetchall()
            print(f'{data}')
            cursor.close()
            break
        except OperationalError:
            time.sleep(2)
        except Exception as ex:  # pylint: disable=broad-except
            print(f'{ex.__module__}.{ex.__class__.__name__}')
            time.sleep(0.5)


@task
def wait_for_redis(ctx, service_name='redis'):
    from redis import Redis
    from redis.exceptions import RedisError
    import yaml
    with open('docker-compose.yml', 'r') as f:
        compose = yaml.load(f, Loader=yaml.SafeLoader)
        redis_port = compose['services'][service_name]['ports'][0].split(':')[1]
    n_max = 40
    for x in range(1, n_max + 1):
        try:
            print(f'checking for redis on port {redis_port} attempt #{x}')
            r = Redis(host='127.0.0.1', port=redis_port)
            r.get('hello')
            break
        except RedisError:
            print('error connecting to redis')
            time.sleep(0.5)



from raft import task

from convocations.aws.base import AwsTask


@task(klass=AwsTask)
def start_task(ctx, task_name, command, cluster=None, tail=True, session=None, **kwargs):
    """
    kicks off a task in ecs

    ```yaml
    tundra:
      ecs:
        cluster: cluster1
        network_configuration:
          subnets:
            - subnet-0123456789
            - subnet-1234567890
          security_groups:
            - sg-0123456789
          assignPublicIp: DISABLED
    ```
    :param ctx:
    :param task_name: the name of the task
    :param command: the command to run inside the container
    :param cluster: the cluster name
    :param tail: if specified, we will tail the logs
    :param session:
    :param kwargs:

    :return:
    """
    from convocations.aws.ecs import get_latest_revision
    from convocations.aws.ecs import task_logs
    from convocations.base.utils import get_context_value, notice, notice_end
    x = get_latest_revision(ctx, task_name, session=session, quiet=True)
    ecs = session.client('ecs')
    command = command.split(' ')
    cluster = cluster or get_context_value(ctx, 'tundra.ecs.cluster')
    network = get_context_value(ctx, 'tundra.ecs.network_configuration')
    if network:
        network = dict(awsvpcConfiguration=network)
    notice(f'starting {x} on cluster {cluster}')
    result = ecs.run_task(
        cluster=cluster,
        taskDefinition=x,
        overrides={
            'containerOverrides': [{'name': task_name, 'command': command}]
        },
        launchType='FARGATE',
        networkConfiguration=network
    )
    result = result['tasks'][0]
    task_arn = result['taskArn']
    task_id = task_arn.rsplit('/', 1)[-1]
    notice_end(task_id)
    if tail:
        task_logs(ctx, cluster, task_name, task_id)



import os
from collections import defaultdict
import json

from raft import task

from ...base.utils import notice
from ...base.utils import notice_end
from ...base.utils import print_table
from ..base import AwsTask
from ..base import yielder


def list_org_accounts(session):
    notice('getting accounts')
    accounts = list(yielder(
        'organizations', 'list_accounts', session,
        PaginationConfig=dict(MaxItems=100)))
    notice_end(f'{len(accounts)}')
    return accounts


@task(klass=AwsTask)
def org_accounts(ctx, name=None, session=None, **kwargs):
    """
    lists all accounts in the organization in a nice tabular format
    """
    accounts = list_org_accounts(session)
    print()
    # header = [ 'name', 'id', 'email' ]
    header = [ 'name', 'id', ]
    rows = []
    name = (name or '').lower()
    for x in accounts:
        if name in x['Name'].lower():
            rows.append([
                x['Id'],
                x['Name'],
                # x['Email'],
            ])
    print_table(header, rows)


@task(klass=AwsTask)
def sso_groups(ctx, name=None, session=None, **kwargs):
    """
    lists all groups in the aws sso in a nice tabular format
    """
    notice('getting instances')
    client = session.client('sso-admin')
    response = client.list_instances(MaxResults=10)
    notice_end()
    instances = response['Instances']
    if not instances:
        return
    header = [ 'arn', 'identity_store_id', ]
    rows = []
    for x in instances:
        rows.append([
            x['InstanceArn'],
            x['IdentityStoreId'],
        ])
    print('===================')
    print_table(header, rows)
    print('===================')
    identity_store_id = instances[0]['IdentityStoreId']
    identity_store = session.client('identitystore')
    kwargs = dict(IdentityStoreId=identity_store_id)
    groups = list(yielder(identity_store, 'list_groups', **kwargs))
    header = [ 'group_id', 'name', 'external_id' ]
    rows = []
    if name:
        name = name.lower()
    for x in groups:
        display_name = x['DisplayName'].lower()
        if name and name not in display_name:
            continue
        rg = [
            x['GroupId'],
            x['DisplayName'],
        ]
        external_ids = x.get('ExternalIds') or []
        external_id = ''
        if external_ids:
            external_id = external_ids[0]['Id']
        rg.append(external_id)
        rows.append(rg)
    print_table(header, rows)


@task(klass=AwsTask)
def export_permission_sets(ctx, filename, session=None, prefix=None, **kwargs):
    """
    exports permission set to cloud formation
    """
    if not filename.endswith('.yml'):
        filename = f'{filename}.yml'
    notice('getting instance arns')
    sso_admin = session.client('sso-admin')
    instances = sso_admin.list_instances()
    instances = instances['Instances']
    permission_sets = defaultdict(list)
    notice_end(f'{len(instances)}')
    for instance in instances:
        instance_arn = instance['InstanceArn']
        notice(f'permission sets for {instance_arn}')
        g = yielder(
            sso_admin,
            'list_permission_sets',
            InstanceArn=instance_arn)
        rg = list(g)
        notice_end(f'{len(rg)}')
        permission_sets[instance_arn] = rg
    exports = {}
    n = 0
    for instance_arn, permission_set_arns in permission_sets.items():
        for permission_set_arn in permission_set_arns:
            n += 1
            notice(f'exporting {permission_set_arn}')
            permission_set = sso_admin.describe_permission_set(
                PermissionSetArn=permission_set_arn,
                InstanceArn=instance_arn,
            )['PermissionSet']
            managed_policies = sso_admin.list_managed_policies_in_permission_set(
                InstanceArn=instance_arn,
                PermissionSetArn=permission_set_arn,
                MaxResults=100,
            )['AttachedManagedPolicies']
            inline_policy = sso_admin.get_inline_policy_for_permission_set(
                InstanceArn=instance_arn,
                PermissionSetArn=permission_set_arn,
            )['InlinePolicy']
            exported_permission_set = dict(
                InstanceArn=instance_arn,
                Name=permission_set['Name'],
                SessionDuration=permission_set['SessionDuration'],
            )
            description = permission_set.get('Description')
            if description:
                exported_permission_set['Description'] = description
            managed_policies = [ x['Arn'] for x in managed_policies ]
            if managed_policies:
                exported_permission_set['ManagedPolicies'] = managed_policies
            if inline_policy:
                exported_permission_set['InlinePolicy'] = json.loads(inline_policy)
            exports[f'PermissionSet{n:02d}'] = dict(
                Type='AWS::SSO::PermissionSet',
                Properties=exported_permission_set
            )
            notice_end()
    p = os.path.join(prefix, 'cfn', filename)
    with open(p, 'w') as f:
        from convocations.base.utils import yaml_serializer
        template = dict(
            AWSTemplateFormatVersion='2010-09-09',
            Resources=exports,
        )
        yaml_serializer().dump(template, f)


@task(klass=AwsTask)
def export_assignments(ctx, identity_store_id, filename, session=None, prefix=None, **kwargs):
    """
    exports permission set to cloud formation
    """
    if not filename.endswith('.yml'):
        filename = f'{filename}.yml'
    notice('getting instance arns')
    sso_admin = session.client('sso-admin')
    instances = sso_admin.list_instances()
    instances = instances['Instances']
    notice_end(f'{len(instances)}')

    identity_store = session.client('identitystore')
    user_map = {}
    group_map = {}
    accounts = list_org_accounts(session)
    exports = {}
    n = 0
    for instance in instances:
        instance_arn = instance['InstanceArn']
        for account in accounts:
            account_id = account['Id']
            g = yielder(
                sso_admin,
                'list_permission_sets_provisioned_to_account',
                InstanceArn=instance_arn,
                AccountId=account_id,
                PaginationConfig=dict(MaxItems=100))
            for permission_set_arn in g:
                h = yielder(
                    sso_admin,
                    'list_account_assignments',
                    InstanceArn=instance_arn,
                    AccountId=account_id,
                    PermissionSetArn=permission_set_arn,
                    PaginationConfig=dict(PageSize=100))
                for assignment in h:
                    n += 1
                    principal_id = assignment['PrincipalId']
                    principal_type = assignment['PrincipalType']
                    if principal_type == 'GROUP':
                        if principal_id not in group_map:
                            r = identity_store.describe_group(
                                IdentityStoreId=identity_store_id,
                                GroupId=principal_id
                            )
                            group_map[principal_id] = r['DisplayName']
                        principal_id = group_map.get(principal_id)
                    else:
                        if principal_id not in user_map:
                            r = identity_store.describe_user(
                                IdentityStoreId=identity_store_id,
                                UserId=principal_id
                            )
                            user_map[principal_id] = r['UserName']
                        principal_id = user_map[principal_id]
                    props = dict(
                        InstanceArn=instance_arn,
                        TargetType='AWS_ACCOUNT',
                        TargetId=account_id,
                        PermissionSetArn=permission_set_arn,
                        PrincipalId=principal_id,
                        PrincipalType=principal_type,
                    )
                    exports[f'Assignment{n:03}'] = dict(
                        Type='AWS::SSO::Assignment',
                        Properties=props,
                    )

    p = os.path.join(prefix, 'cfn', filename)
    with open(p, 'w') as f:
        from convocations.base.utils import yaml_serializer
        template = dict(
            AWSTemplateFormatVersion='2010-09-09',
            Resources=exports,
        )
        yaml_serializer().dump(template, f)



from raft import task
from ..base import AwsTask


@task(klass=AwsTask)
def route_table_to_yaml(ctx, logical_name=None, route_table_id=None,
                        session=None, quiet=False, **kwargs):
    """
    generates the cloudformation yaml for a security group
    """
    ec2 = session.resource('ec2')
    table = ec2.RouteTable(route_table_id)
    properties = dict(
        VpcId=table.vpc_id,
        Tags=table.tags,
    )
    properties = { k: v for k, v in properties.items() if v is not None }
    data = {
        logical_name: dict(
            Type='AWS::EC2::RouteTable',
            DeletionPolicy='Retain',
            Properties=properties,
        ),
    }
    if not quiet:
        from convocations.base.utils import dump_yaml
        dump_yaml(data)
    return data


@task(klass=AwsTask)
def import_route_table(ctx, logical_name, route_table_id, filename,
                       skip_review=False, session=None, prefix=None,
                       **kwargs):
    """
    creates a yaml template for a route table and imports it into cloudformation

    if skip_review is set, we will immediately try to execute the change set
      in cloudformation
    """
    from .import_resource_to_cfn import import_resource_to_cfn
    data = route_table_to_yaml(
        ctx, logical_name, route_table_id, session=session, quiet=True)
    resources_to_import = [dict(
        ResourceType='AWS::EC2::RouteTable',
        LogicalResourceId=list(data.keys())[0],
        ResourceIdentifier=dict(RouteTableId=route_table_id),
    )]
    import_resource_to_cfn(
        data, resources_to_import,
        filename, None, skip_review, session, prefix)



import os
from datetime import datetime
from datetime import timedelta
from tempfile import NamedTemporaryFile
import time
import boto3
from raft.tasks import task
from .ec2 import instance_by_name, aws_filter
from .base import AwsTask, get_account_id, get_tag
from ..base.utils import notice, notice_end, confirm
from ..base.utils import print_table


def get_latest_ami(name, session):
    ec2 = session.client('ec2')
    responses = ec2.describe_images(Filters=[{
        'Name': 'name',
        'Values': [ f'*{name}*', ],
    }], Owners=[
        session.account_id,
    ])
    images = responses['Images']
    dt = datetime.utcnow()
    latest_image = None
    for image in images:
        creation_date = image['CreationDate']
        creation_date = datetime.strptime(creation_date[:-5], '%Y-%m-%dT%H:%M:%S')
        if dt - creation_date < timedelta(minutes=10):
            latest_image = image
    return latest_image


def load_image(image, session):
    ec2 = session.resource('ec2')
    image_id = image['ImageId']
    return ec2.Image(image_id)


def public_ip(ami_name, session=None):
    """
    Grabs the public IP address from the output in
    cloudformation
    """
    session = session or boto3.DEFAULT_SESSION
    cfn = session.resource('cloudformation')
    notice('getting ip')
    stack = cfn.Stack(f'{ami_name}-ami')
    stack.load()
    outputs = stack.outputs or []
    result = None
    for x in outputs:
        if x['OutputKey'] == 'Ip':
            result = x['OutputValue']
            break
    notice_end(result or 'not found')
    return result


def tail_output_log(ctx, ip, name, session):
    """
    In our standard ami builds, the
    user data log file will be written to
    /var/log/{ami_name}-output.log.  Tail it here
    """
    wait_time = 30
    username = 'ubuntu'
    if 'centos' in name:
        username = 'centos'
        wait_time = 90
    elif 'oracle' in name:
        username = 'ec2-user'
        wait_time = 90
    elif 'amazon' in name or 'amzn' in name:
        username = 'ec2-user'
        wait_time = 20
    instance = instance_by_name(name, session)
    key_pair = get_tag(instance, 'key_name') or instance.key_name
    if not key_pair.startswith('/'):
        key_pair = f'/{key_pair}'
    notice(f'loading {key_pair} from ssm')
    ssm = session.client('ssm')
    x = ''
    local_key = None
    try:
        try:
            key = ssm.get_parameter(Name=key_pair, WithDecryption=True)
            with NamedTemporaryFile('w', suffix='.key', delete=False) as f:
                f.write(key['Parameter']['Value'])
                local_key = f.name
            os.chmod(local_key, 0o600)
            x = f'-i {local_key}'
            notice_end()
        except Exception as ex:
            notice_end(f'{ex}')
        notice(f'waiting {wait_time} seconds for ssh')
        time.sleep(wait_time)
        notice_end()

        ctx.run(
            f'ssh {x} -o "ServerAliveInterval 2" {username}@{ip}'
            f' "sudo tail -f /var/log/{name}-ami-output.log | '
            f' stdbuf -o0 sed \'/^Cloud-init.*finished at.*/ q\'"', warn=True)
        notice_end('----- end of output log -----')
    finally:
        if os.path.exists(local_key):
            os.remove(local_key)


@task(klass=AwsTask)
def wait_for_ami(ctx, name, session=None, **kwargs):
    """
    waits for an ami with `name` in it
    """
    # first check if we have a public ip available
    ip = public_ip(name, session)
    if ip:
        tail_output_log(ctx, ip, name, session)
    name = name[:-4]
    p_image = None
    for i in range(1, 41):
        if p_image:
            notice(f'checking {p_image.id}/{i}')
            p_image.load()
            notice_end(p_image.state)
            if p_image.state == 'available':
                break
        else:
            notice(f'check #{i}')
            latest_image = get_latest_ami(name, session)
            if not latest_image:
                notice_end('not yet initiated')
            else:
                p_image = load_image(latest_image, session)
                notice_end(f'{p_image.id}/{p_image.state}')
                if p_image.state == 'available':
                    break
        time.sleep(15)


@task(positional=['name'], klass=AwsTask, help={
    'name': 'the name of the ami to look for, we will search case-insensitive '
            'for partial match',
    'owners': 'a comma delimited list of owner ami ids to filter by',
    'product_codes': 'a comma delimited list of product codes to filter by',
})
def amis(ctx, name=None, owners=None, product_codes=None, session=None, profile=None, **kwargs):
    """
    Lists all of our amis in aws
    """
    notice('getting images')
    images = []
    if name:
        name = name.lower()
    ec2 = session.client('ec2')
    owners = owners.split(',') if owners else [ 'self' ]
    filters = None
    if product_codes:
        owners.append('aws-marketplace')
        product_codes = product_codes.split(',')
        filters = [ aws_filter('product-code', x) for x in product_codes ]
    result = ec2.describe_images(Owners=owners, Filters=filters)
    for x in result['Images']:
        if not name or name in x['Name'].lower():
            images.append(x)
    images.sort(key=lambda lx: lx['CreationDate'])
    images = images[::-1]
    rows = []
    for x in images:
        row = [
            x['ImageId'], x['Name'],
            x['CreationDate'].split('T', 1)[0], x['State']
        ]
        snapshots = []
        for device_mapping in x['BlockDeviceMappings']:
            snapshots.append(device_mapping['Ebs']['SnapshotId'])
        row.append(', '.join(snapshots))
        rows.append(row)

    notice_end()
    print_table(['id', 'description', 'at', 'status', 'snapshot_id' ], rows)


@task(klass=AwsTask, help={
    'image_id': 'the ami id of the image to delete',
})
def delete_ami(ctx, image_id, session=None, **kwargs):
    """
    Deletes the ami and any associated snapshots
    """
    account_id = get_account_id(session)
    ec2 = session.resource('ec2')
    notice('loading image')
    image = ec2.Image(image_id)
    image.load()
    if not image.meta.data:
        notice_end('not found')
        return
    notice_end(image.name)
    notice('finding snapshot')
    snapshot = None
    client = boto3.client('ec2')
    data = client.describe_snapshots(OwnerIds=[ account_id ], MaxResults=100)
    snapshots = data['Snapshots']
    for x in snapshots:
        description = x['Description']
        if image_id in description:
            snapshot = x
            break
    if snapshot:
        snapshot_id = snapshot['SnapshotId']
        snapshot = ec2.Snapshot(snapshot_id)
        notice_end(f'{snapshot.id}')
    else:
        notice_end('n/a')
    r = confirm(f'You are deleting ami [{image.name}]')
    if not r:
        return

    notice('deregistering ami')
    image.deregister()
    notice_end()
    if snapshot:
        notice('deleting snapshot')
        snapshot.delete()
        notice_end()


@task(klass=AwsTask)
def centos_amis(ctx, session=None, prefix=None, **kwargs):
    """
    finds all marketplace amis released under the centos product code
    """
    amis(ctx, product_codes='aw0evgkw8e5c1q413zgy5pjce', session=session)


@task(klass=AwsTask)
def oracle_amis(ctx, session=None, prefix=None, **kwargs):
    """
    finds all marketplace amis released under the oracle owner id
    """
    amis(ctx, owners='131827586825', session=session, prefix=prefix)


@task(klass=AwsTask)
def ubuntu_amis(ctx, session=None, prefix=None, **kwargs):
    """
    finds all marketplace amis released under the ubuntu owner id
    """
    amis(ctx, owners='099720109477', session=session, prefix=prefix)




import warnings

from raft import Task
from raft.context import Context
from ..aws.base import make_session


try:
    from urllib3.exceptions import InsecureRequestWarning
    import requests
    import yaml
    from murmuration import kms_wrapped
    from lxml import etree
except ImportError:
    pass


class XmlApiTask(Task):
    def __call__(self, *args, **kwargs):
        ctx = args[0]
        host = kwargs.get('host')
        profile = kwargs.get('profile')
        bucket = kwargs.get('bucket')
        session = kwargs.get('session')
        if isinstance(ctx, Context):
            if profile is None:
                try:
                    profile = ctx.aws.profile
                except AttributeError:
                    pass
            if host is None:
                try:
                    host = ctx.palo_alto.host
                    kwargs['host'] = host
                except AttributeError:
                    pass
            if bucket is None:
                try:
                    bucket = ctx.palo_alto.bucket
                    kwargs['bucket'] = bucket
                except AttributeError:
                    pass
            kwargs['username'] = ctx.palo_alto.username
            kwargs['password'] = ctx.palo_alto.password
        if host and 'username' in kwargs and 'password' in kwargs:
            api = XmlApi(**kwargs)
            kwargs['api'] = api
        session = make_session(profile, session)
        kwargs['session'] = session
        super().__call__(*args, **kwargs)


class XmlApi:
    def __init__(self, host, username=None, password=None, profile=None, **kwargs):
        self.session = requests.Session()
        self.session.verify = False
        self.host = host
        self.base_url = f'https://{self.host}/api/'
        self.generate(host, username, password, profile, **kwargs)

    def generate(self, host, username=None, password=None, profile=None, **kwargs):
        """
        generates an api key using the rest api with the palo alto
        """
        filename = kwargs.get('filename')
        if filename:
            with open(filename, 'r') as f:
                conf = yaml.load(f, Loader=yaml.SafeLoader)
            api_conf = conf.get('api') or {}
            username = username or api_conf.get('username')
            password = password or api_conf.get('password')
            profile = profile or conf.get('profile')
        if password:
            try:
                password = kms_wrapped.decrypt(password, profile=profile)
            except:
                pass
        doc = self.post('keygen', None, user=username, password=password)
        api_key = doc.text
        self.session.headers = {
            'X-PAN-KEY': api_key,
        }

    def raw(self, method, params=None, data=None, files=None):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=InsecureRequestWarning)
            response = self.session.request(
                method, self.base_url,
                params=params, data=data, files=files)
            return response

    def get(self, type_, action=None, **kwargs):
        params = {}
        params.update(kwargs)
        params['type'] = type_
        if action:
            params['action'] = action
        response = self.raw('get', params)
        return self.handle_response(response)

    @classmethod
    def handle_response(cls, response):
        doc = etree.fromstring(response.content)
        if doc.tag == 'response':
            stuff = doc.xpath('/response/result')
            if stuff:
                stuff = stuff[0]
            for x in stuff:
                return x
        print('returning doc as is')
        return doc

    def post(self, type_, action, **kwargs):
        params = {}
        params.update(kwargs)
        params['type'] = type_
        if action:
            params['action'] = action
        response = self.raw('post', data=params)
        return self.handle_response(response)

    def import_cert(self, name, category, filename, **kwargs):
        params = {}
        params.update(kwargs)
        params['type'] = 'import'
        params['category'] = category
        params['certificate-name'] = name
        params['format'] = 'pem'
        files = {}
        with open(filename, 'r') as f:
            files['file'] = f.read()
        response = self.raw('post', data=params, files=files or None)
        return self.handle_response(response)



from imaplib import IMAP4_SSL as Imap
from raft import task


@task
def test_imap(ctx, user, filename='conf/dev.yml'):
    """
    tests imap connectivity via oauth to office 365
    """
    from requests_oauthlib import OAuth2Session
    from oauthlib.oauth2 import BackendApplicationClient
    from waddle import load_config
    conf = load_config(filename)
    client_id = conf.imap.client_id
    client_secret = conf.imap.client_secret
    scope = ' '.join(conf.imap.scopes)
    token_url = conf.imap.token_url
    client = BackendApplicationClient(
        client_id=client_id,
        scope=scope,
    )
    session = OAuth2Session(client=client)
    session.fetch_token(
        token_url,
        scope=scope,
        client_secret=client_secret)
    access_token = session.access_token
    auth = f'user={user}\1auth=Bearer {access_token}\1\1'.encode('utf-8')
    mbox = Imap('outlook.office365.com', 993)
    mbox.debug = 4
    mbox.authenticate('XOAUTH2', lambda lx: auth)
    mbox.select('Inbox', True)
    mbox.close()



from raft.tasks import task

from ...base.utils import notice, notice_end, print_table
from ..base import AwsTask
from ..base import yielder


@task(klass=AwsTask)
def nameservers(ctx, zone_name, session=None, **kwargs):
    """
    lists all of the name servers for a particular zone name
    """
    route53 = session.client('route53')
    g = yielder(route53, 'list_hosted_zones', session)
    header = [ 'zone', 'id', 'nameserver' ]
    rows = []
    for x in g:
        if zone_name.lower() in x['Name']:
            notice(x['Name'])
            hosted_zone = route53.get_hosted_zone(Id=x['Id'])
            if 'DelegationSet' in hosted_zone:
                ns = hosted_zone['DelegationSet']['NameServers']
                for nameserver in ns:
                    rows.append([ x['Name'], x['Id'], nameserver ])
            else:
                rows.append([ x['Name'], x['Id'], '' ])
            notice_end()
    print_table(header, rows)



import os
from collections import OrderedDict
from uuid import uuid4
from ...base.utils import notice, notice_end


def to_pascal_case(m, exclude=None):
    """
    the cfn for many ecs resources closely resembles the
    structure needed by cloudformation.  we can easily convert
    api format to cloudformation format by pascal casing the
    values we get back from the api.

    recursively works across lists and dicts as needed to
    pascal case all keys.  Returns scalars as-is.

    e.g., roleArn => RoleArn
    """
    if not m:
        return None
    if isinstance(m, (dict, OrderedDict)):
        result = {}
        for key, value in m.items():
            key = f'{key[0].upper()}{key[1:]}'
            if isinstance(value, (dict, OrderedDict, list)):
                if not exclude or key not in exclude:
                    value = to_pascal_case(value, exclude)
                if value is not None:
                    result[key] = value
            else:
                result[key] = value
        return result or None
    if isinstance(m, list):
        result = [ to_pascal_case(x, exclude) for x in m ]
        return result or None
    return m


def cfn_filename(prefix, stack_name):
    filename = stack_name or f'{stack_name}.yml'
    if not filename.endswith('.yml'):
        filename = f'{filename}.yml'
    stack_name, _ = os.path.splitext(filename)
    filepath = os.path.join(prefix, 'cfn', filename)
    return filepath, stack_name


def update_template(filepath, data):
    from ruamel.yaml import load, RoundTripLoader
    from ...base.utils import yaml_serializer
    notice(f'updating {filepath}')
    if os.path.exists(filepath):
        with open(filepath, 'r') as f:
            template_body = load(f, RoundTripLoader)
        template_body['Resources'].update(data)
    else:
        template_body = dict(
            AWSTemplateFormatVersion='2010-09-09',
            Description='imported via upholstery',
            Resources=data,
        )
        notice_end('not found')
    with open(filepath, 'w') as f:
        yaml_serializer().dump(template_body, f)
    notice_end()


def import_resource_to_cfn(
        data, resources_to_import,
        filename, name, skip_review=False,
        session=None, prefix=None):
    from .stack_commands import wait_for_event
    from .stack_commands import push_template_to_s3
    filepath, stack_name = cfn_filename(prefix, filename or name)
    update_template(filepath, data)
    url, template_body = push_template_to_s3(stack_name, filepath, session)
    cfn = session.client('cloudformation')
    cfn_resource = session.resource('cloudformation')
    kwargs = {}
    try:
        stack = cfn_resource.Stack(stack_name)
        stack.load()
        kwargs['Parameters'] = [
            dict(
                ParameterKey=param['ParameterKey'],
                UsePreviousValue=True)
            for param in stack.parameters ]
    except:
        pass
    change_set_name = f'{stack_name}-{uuid4().hex}'
    notice('creating change set')
    if url:
        kwargs['TemplateURL'] = url
    else:
        kwargs['TemplateBody'] = template_body
    cfn.create_change_set(
        StackName=stack_name,
        ChangeSetName=change_set_name,
        ChangeSetType='IMPORT',
        ResourcesToImport=resources_to_import,
        **kwargs
    )
    notice_end(change_set_name)
    if skip_review:
        waiter = cfn.get_waiter('change_set_create_complete')
        kwargs = dict(
            ChangeSetName=change_set_name,
            StackName=stack_name
        )
        waiter.wait(WaiterConfig=dict(Delay=1), **kwargs)
        notice(f'executing {change_set_name}')
        cfn.execute_change_set(**kwargs)
        notice_end()
        wait_for_event(session, stack_name, 'stack_import_complete', None)


def sync_resource(prefix, filename, data):
    """
    fixes drift in a template by updating one resource with the data provided
    """
    filepath, _ = cfn_filename(prefix, filename)
    update_template(filepath, data)


def filtered_tags(obj):
    """
    removes cloudformation tags from the tags array
    """
    tags = obj.get('tags') or []
    return [ x for x in tags if x['key'][:4] != 'aws:' ]



# helpers for use with aws resource access manager
from raft.tasks import task
from ...base.utils import notice
from ...base.utils import notice_end
from ..base import AwsTask


__all__ = [
    'accept_sharing_invitation',
]


@task(klass=AwsTask)
def accept_sharing_invitation(ctx, name, session=None, **kwargs):
    """
    Accepts the sharing invitation with the provided name

    n.b., the first invitation containing name will be accepted
          the comparison is case-insensitive
    """
    ram = session.client('ram')
    notice('looking for invitations')
    response = ram.get_resource_share_invitations()
    results = response.get('resourceShareInvitations')
    for x in results:
        if x['status'] != 'PENDING':
            continue
        invitation_name = x['resourceShareName'].lower()
        notice('name')
        notice_end(invitation_name)
        if name.lower() in invitation_name:
            notice_end(f'{invitation_name}')
            notice(f'accepting {invitation_name}')
            arn = x['resourceShareInvitationArn']
            ram.accept_resource_share_invitation(
                resourceShareInvitationArn=arn
            )
            notice_end()
            return
    if not results:
        notice_end('no invitations found')
        return
    notice_end()



from botocore.exceptions import WaiterError
from ..base.utils import notice
from ..base.utils import notice_end


def wait(waiter, message, **kwargs):
    config = kwargs.pop('WaiterConfig', None)
    config = config or {}
    config.setdefault('Delay', 15)
    config.setdefault('MaxAttempts', 2)
    for n in range(1, 41):
        notice(f'waiting for {message} {n}/40')
        try:
            waiter.wait(WaiterConfig=config, **kwargs)
        except WaiterError:
            notice_end('not yet')
        else:
            notice_end()
            break



from raft.tasks import task
from ..base.projects import new_project as base_new_project


@task(help={
    'name': 'the name of the new project, a new directory '
            'called <name> will be created in <dir_name>',
    'dir_name': 'the /path/to the directory in which the new '
                'project will be created',
    'setup': 'if specified, we will try to run the setup steps '
             'after creating the directory from the project template'
})
def new_project(ctx, name, dir_name='.', setup=False):
    """
    starts a new project for a docker container using the slytherin standard template

    this will include the following:
      * setup and build via Makefile / make or convocations
      * Dockerfile
      * publishing to docker.io as default
    """
    jinja_templates = [
        'README.md',
        'tasks/build.py',
        'templates/buildspec.yml',
    ]
    base_new_project(
        ctx, name, dir_name, setup=setup,
        jinja_templates=jinja_templates, package=__package__)



import csv
from raft.tasks import task

from ..base.utils import notice, notice_end
from .base import OneLoginTask


@task(klass=OneLoginTask)
def user_report(ctx, filename, api=None):
    """
    saves a csv of the users, active status, and last login time to a csv
    """
    notice('getting onelogin users')
    header = [ 'firstname', 'lastname', 'email', 'state', 'status', 'last_login', 'role_ids' ]
    roles = api.client.get_roles()
    role_map = { x.id: x.name.lower() for x in roles }
    users = api.client.get_users(query_parameters=dict(
        fields=','.join(header),
    ))
    notice_end(len(users))
    rows = []
    for x in users:
        row = [
            x.firstname,
            x.lastname,
            x.email,
            f'{x.state}',
            f'{x.status}',
            x.last_login.isoformat() if x.last_login else '',
            ';'.join(map(lambda lx: role_map[lx], x.role_ids)),
        ]
        rows.append(row)
    with open(filename, 'w') as f:
        writer = csv.writer(f, csv.excel)
        writer.writerow(header)
        writer.writerows(rows)



from raft.collection import Collection
from .layers import layers
from .objects import group
from .services import services, service_group
from .services import service
from .versions import api_version


ns = Collection(
    layers,
    services,
    api_version,
    service,
    service_group,
    group,
)



from raft import task
from .base import XmlApiTask, XmlApi


try:
    from lxml import etree
except ImportError:
    pass


@task(klass=XmlApiTask)
def save_config(ctx, host=None, bucket=None, config=None, **kwargs):
    """
    saves off the running config from the specified hosts.
    if bucket is specified, the config is saved as the bootstrap
    config in the specified bucket
    """
    api = kwargs.get('api')
    api = api or XmlApi(host, filename=config)
    doc = api.get('config', 'show')
    if not bucket:
        print(etree.tostring(doc, pretty_print=True).decode('utf-8'))
    else:
        from ..aws.base import make_session
        session = kwargs.get('session')
        session = make_session(kwargs.get('profile'), session)
        s3 = session.client('s3')
        s3.put_object(
            Bucket=bucket,
            Key='config/bootstrap.xml',
            Body=etree.tostring(doc, pretty_print=True))


@task(klass=XmlApiTask)
def export_config(ctx, host=None, bucket=None, config=None, **kwargs):
    """
    exports the running config from the specified hosts.
    if bucket is specified, the config is saved as the bootstrap
    config in the specified bucket
    """
    api = XmlApi(host, filename=config)
    doc = api.get('export', category='configuration', )
    if not bucket:
        print(etree.tostring(doc, pretty_print=True).decode('utf-8'))
    else:
        from ..aws.base import make_session
        session = kwargs.get('session')
        session = make_session(kwargs.get('profile'), session)
        s3 = session.client('s3')
        s3.put_object(
            Bucket=bucket,
            Key='config/bootstrap.xml',
            Body=etree.tostring(doc, pretty_print=True))
